@@ -1,3 +1,17 @@
+#  Copyright 2022 Red Hat, Inc.
+#
+#  Licensed under the Apache License, Version 2.0 (the "License");
+#  you may not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing, software
+#  distributed under the License is distributed on an "AS IS" BASIS,
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#  See the License for the specific language governing permissions and
+#  limitations under the License.
+
 import asyncio
 import concurrent.futures
 import glob
@@ -15,18 +29,15 @@ from pprint import pprint
 from typing import Callable, Dict, List, Optional, Union
 
 import ansible_runner
-import dpath.util
+import dpath
 import janus
 import yaml
-
-if os.environ.get("RULES_ENGINE", "drools") == "drools":
-    from drools import ruleset as lang
-else:
-    from durable import lang
+from drools import ruleset as lang
 
 from .collection import find_playbook, has_playbook, split_collection_name
 from .conf import settings
 from .exception import ShutdownException
+from .job_template_runner import job_template_runner
 from .util import get_horizontal_rule
 
 logger = logging.getLogger(__name__)
@@ -41,23 +52,24 @@ async def none(
     variables: Dict,
     facts: Dict,
     project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
 ):
     await event_log.put(
         dict(
             type="Action",
             action="noop",
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
             activation_id=settings.identifier,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
         )
     )
 
 
 async def debug(event_log, **kwargs):
-    print(get_horizontal_rule("="))
-    print(get_horizontal_rule("="))
-    print("facts:")
-    pprint(lang.get_facts(kwargs["ruleset"]))
     print(get_horizontal_rule("="))
     print("kwargs:")
     pprint(kwargs)
@@ -68,8 +80,11 @@ async def debug(event_log, **kwargs):
             type="Action",
             action="debug",
             playbook_name=kwargs.get("name"),
+            ruleset=kwargs.get("source_ruleset_name"),
+            rule=kwargs.get("source_rule_name"),
             activation_id=settings.identifier,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(kwargs.get("variables")),
         )
     )
 
@@ -81,6 +96,8 @@ async def print_event(
     variables: Dict,
     facts: Dict,
     project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
     name: Optional[str] = None,
     var_root: Union[str, Dict, None] = None,
@@ -104,8 +121,44 @@ async def print_event(
             type="Action",
             action="print_event",
             activation_id=settings.identifier,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
             playbook_name=name,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
+        )
+    )
+
+
+async def echo(
+    event_log,
+    inventory: Dict,
+    hosts: List,
+    variables: Dict,
+    facts: Dict,
+    project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
+    ruleset: str,
+    name: Optional[str] = None,
+    var_root: Union[str, Dict, None] = None,
+    message: Optional[str] = None,
+):
+    if var_root:
+        update_variables(variables, var_root)
+
+    print(str(datetime.now()) + " : " + message)
+    sys.stdout.flush()
+    await event_log.put(
+        dict(
+            type="Action",
+            action="echo",
+            activation_id=settings.identifier,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
+            playbook_name=name,
+            run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
         )
     )
 
@@ -117,6 +170,8 @@ async def set_fact(
     variables: Dict,
     facts: Dict,
     project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
     fact: Dict,
     name: Optional[str] = None,
@@ -128,8 +183,11 @@ async def set_fact(
             type="Action",
             action="set_fact",
             activation_id=settings.identifier,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
             playbook_name=name,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
         )
     )
 
@@ -141,6 +199,8 @@ async def retract_fact(
     variables: Dict,
     facts: Dict,
     project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
     fact: Dict,
     name: Optional[str] = None,
@@ -150,9 +210,12 @@ async def retract_fact(
         dict(
             type="Action",
             action="retract_fact",
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
             activation_id=settings.identifier,
             playbook_name=name,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
         )
     )
 
@@ -162,8 +225,10 @@ async def post_event(
     inventory: Dict,
     hosts: List,
     variables: Dict,
-    project_data_file: str,
     facts: Dict,
+    project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
     event: Dict,
 ):
@@ -173,8 +238,11 @@ async def post_event(
         dict(
             type="Action",
             action="post_event",
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
             activation_id=settings.identifier,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
         )
     )
 
@@ -186,6 +254,8 @@ async def run_playbook(
     variables: Dict,
     facts: Dict,
     project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
     name: str,
     set_facts: Optional[bool] = None,
@@ -201,12 +271,11 @@ async def run_playbook(
 ):
 
     logger.info("running Ansible playbook: %s", name)
-    temp, playbook_name, job_id = await pre_process_runner(
+    temp_dir, playbook_name = await pre_process_runner(
         event_log,
         inventory,
         variables,
         facts,
-        ruleset,
         name,
         "run_playbook",
         var_root,
@@ -215,13 +284,31 @@ async def run_playbook(
         project_data_file,
         **kwargs,
     )
+
+    job_id = str(uuid.uuid4())
+
+    logger.info(f"ruleset: {source_ruleset_name}, rule: {source_rule_name}")
+    await event_log.put(
+        dict(
+            type="Job",
+            job_id=job_id,
+            ansible_rulebook_id=settings.identifier,
+            name=playbook_name,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
+            hosts=",".join(hosts),
+            action="run_playbook",
+        )
+    )
+
     logger.info("Calling Ansible runner")
 
     if retry:
         retries = max(retries, 1)
     for i in range(retries + 1):
-        if i > 0 and delay > 0:
-            await asyncio.sleep(delay)
+        if i > 0:
+            if delay > 0:
+                await asyncio.sleep(delay)
             logger.info(
                 "Previous run_playbook failed. Retry %d of %d", i, retries
             )
@@ -230,19 +317,21 @@ async def run_playbook(
         await call_runner(
             event_log,
             job_id,
-            temp,
+            temp_dir,
             dict(playbook=playbook_name),
             hosts,
             verbosity,
             json_mode,
         )
-        if get_status(temp) != "failed":
+        if _get_latest_artifact(temp_dir, "status") != "failed":
             break
 
-    return await post_process_runner(
+    await post_process_runner(
         event_log,
-        temp,
+        variables,
+        temp_dir,
         ruleset,
+        source_rule_name,
         settings.identifier,
         name,
         "run_playbook",
@@ -252,6 +341,8 @@ async def run_playbook(
         post_events,
     )
 
+    shutil.rmtree(temp_dir)
+
 
 async def run_module(
     event_log,
@@ -259,6 +350,9 @@ async def run_module(
     hosts: List,
     variables: Dict,
     facts: Dict,
+    project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
     name: str,
     set_facts: Optional[bool] = None,
@@ -274,32 +368,48 @@ async def run_module(
     **kwargs,
 ):
 
-    temp, module_name, job_id = await pre_process_runner(
+    temp_dir, module_name = await pre_process_runner(
         event_log,
         inventory,
         variables,
         facts,
-        ruleset,
         name,
         "run_module",
         var_root,
         copy_files,
         False,
+        project_data_file,
         **kwargs,
     )
+    job_id = str(uuid.uuid4())
+
+    await event_log.put(
+        dict(
+            type="Job",
+            job_id=job_id,
+            ansible_rulebook_id=settings.identifier,
+            name=module_name,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
+            hosts=",".join(hosts),
+            action="run_module",
+        )
+    )
+
     logger.info("Calling Ansible runner")
     module_args_str = ""
     if module_args:
         for k, v in module_args.items():
             if len(module_args_str) > 0:
                 module_args_str += " "
-            module_args_str += f'{k}="{v}"'
+            module_args_str += f"{k}={v!r}"
 
     if retry:
         retries = max(retries, 1)
     for i in range(retries + 1):
-        if i > 0 and delay > 0:
-            await asyncio.sleep(delay)
+        if i > 0:
+            if delay > 0:
+                await asyncio.sleep(delay)
             logger.info(
                 "Previous run_module failed. Retry %d of %d", i, retries
             )
@@ -307,7 +417,7 @@ async def run_module(
         await call_runner(
             event_log,
             job_id,
-            temp,
+            temp_dir,
             dict(
                 module=module_name,
                 host_pattern=",".join(hosts),
@@ -317,13 +427,15 @@ async def run_module(
             verbosity,
             json_mode,
         )
-        if get_status(temp) != "failed":
+        if _get_latest_artifact(temp_dir, "status") != "failed":
             break
 
-    return await post_process_runner(
+    await post_process_runner(
         event_log,
-        temp,
+        variables,
+        temp_dir,
         ruleset,
+        source_rule_name,
         settings.identifier,
         name,
         "run_module",
@@ -332,6 +444,7 @@ async def run_module(
         set_facts,
         post_events,
     )
+    shutil.rmtree(temp_dir)
 
 
 async def call_runner(
@@ -357,7 +470,6 @@ async def call_runner(
     def event_callback(event, *args, **kwargs):
         event["job_id"] = job_id
         event["ansible_rulebook_id"] = settings.identifier
-        logger.debug("event_callback")
         queue.sync_q.put(dict(type="AnsibleEvent", event=event))
 
     # Here we read the async side and push it into the event queue
@@ -421,7 +533,6 @@ async def pre_process_runner(
     inventory: Dict,
     variables: Dict,
     facts: Dict,
-    ruleset: str,
     name: str,
     action: str,
     var_root: Union[str, Dict, None] = None,
@@ -448,19 +559,20 @@ async def pre_process_runner(
     project_dir = os.path.join(private_data_dir, "project")
 
     playbook_name = name
-    if True:
-        os.mkdir(env_dir)
-        with open(os.path.join(env_dir, "extravars"), "w") as f:
-            f.write(yaml.dump(variables))
-        os.mkdir(inventory_dir)
-        with open(os.path.join(inventory_dir, "hosts"), "w") as f:
-            f.write(yaml.dump(inventory))
-        os.mkdir(project_dir)
+
+    os.mkdir(env_dir)
+    with open(os.path.join(env_dir, "extravars"), "w") as f:
+        f.write(yaml.dump(variables))
+    os.mkdir(inventory_dir)
+    with open(os.path.join(inventory_dir, "hosts"), "w") as f:
+        f.write(yaml.dump(inventory))
+    os.mkdir(project_dir)
 
     logger.debug("project_data_file: %s", project_data_file)
     if project_data_file:
         if os.path.exists(project_data_file):
             await untar_project(project_dir, project_data_file)
+            return (private_data_dir, playbook_name)
 
     if check_files:
         if os.path.exists(name):
@@ -483,20 +595,15 @@ async def pre_process_runner(
                 "Could not find a playbook for %s from %s", name, os.getcwd()
             )
 
-    job_id = str(uuid.uuid4())
-
-    await event_log.put(
-        dict(
-            type="Job", job_id=job_id, ansible_rulebook_id=settings.identifier
-        )
-    )
-    return (private_data_dir, playbook_name, job_id)
+    return (private_data_dir, playbook_name)
 
 
 async def post_process_runner(
     event_log,
+    variables: Dict,
     private_data_dir: str,
     ruleset: str,
+    rule: str,
     activation_id: str,
     name: str,
     action: str,
@@ -506,13 +613,8 @@ async def post_process_runner(
     post_events: Optional[bool] = None,
 ):
 
-    for rc_file in glob.glob(
-        os.path.join(private_data_dir, "artifacts", "*", "rc")
-    ):
-        with open(rc_file, "r") as f:
-            rc = int(f.read())
-
-    status = get_status(private_data_dir)
+    rc = int(_get_latest_artifact(private_data_dir, "rc"))
+    status = _get_latest_artifact(private_data_dir, "status")
 
     result = dict(
         type="Action",
@@ -521,17 +623,20 @@ async def post_process_runner(
         playbook_name=name,
         job_id=job_id,
         ruleset=ruleset,
+        rule=rule,
         rc=rc,
         status=status,
         run_at=run_at,
+        matching_events=_get_events(variables),
     )
     await event_log.put(result)
 
     if set_facts or post_events:
         logger.debug("set_facts")
-        for host_facts in glob.glob(
-            os.path.join(private_data_dir, "artifacts", "*", "fact_cache", "*")
-        ):
+        fact_folder = _get_latest_artifact(
+            private_data_dir, "fact_cache", False
+        )
+        for host_facts in glob.glob(os.path.join(fact_folder, "*")):
             with open(host_facts) as f:
                 fact = json.loads(f.read())
             logger.debug("fact %s", fact)
@@ -540,7 +645,125 @@ async def post_process_runner(
             if post_events:
                 lang.post(ruleset, fact)
 
-    return result
+
+async def run_job_template(
+    event_log,
+    inventory: Dict,
+    hosts: List,
+    variables: Dict,
+    facts: Dict,
+    project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
+    ruleset: str,
+    name: str,
+    organization: str,
+    job_args: Optional[dict] = None,
+    set_facts: Optional[bool] = None,
+    post_events: Optional[bool] = None,
+    verbosity: int = 0,
+    var_root: Union[str, Dict, None] = None,
+    copy_files: Optional[bool] = False,
+    json_mode: Optional[bool] = False,
+    retries: Optional[int] = 0,
+    retry: Optional[bool] = False,
+    delay: Optional[int] = 0,
+    **kwargs,
+):
+
+    logger.info(
+        "running job template: %s, organization: %s", name, organization
+    )
+    logger.info("ruleset: %s, rule %s", source_ruleset_name, source_rule_name)
+
+    hosts_limit = ",".join(hosts)
+    if not job_args:
+        job_args = {}
+    job_args["limit"] = hosts_limit
+
+    job_args["extra_vars"] = job_args.get("extra_vars", {})
+    for key in ["fact", "facts", "event", "events"]:
+        if key in variables:
+            job_args["extra_vars"][key] = variables[key]
+
+    job_id = str(uuid.uuid4())
+
+    await event_log.put(
+        dict(
+            type="Job",
+            job_id=job_id,
+            ansible_rulebook_id=settings.identifier,
+            name=name,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
+            hosts=hosts_limit,
+            action="run_job_template",
+        )
+    )
+
+    async def event_callback(event: dict) -> None:
+        await event_log.put(
+            {
+                "type": "AnsibleEvent",
+                "event": {
+                    "uuid": event["uuid"],
+                    "counter": event["counter"],
+                    "stdout": event["stdout"],
+                    "start_line": event["start_line"],
+                    "end_line": event["end_line"],
+                    "event": event["event"],
+                    "created": event["created"],
+                    "parent_uuid": event["parent_uuid"],
+                    "event_data": event["event_data"],
+                    "job_id": job_id,
+                    "controller_job_id": event["job"],
+                    "ansible_rulebook_id": settings.identifier,
+                },
+                "run_at": event["created"],
+            }
+        )
+
+    if retry:
+        retries = max(retries, 1)
+    for i in range(retries + 1):
+        if i > 0:
+            if delay > 0:
+                await asyncio.sleep(delay)
+            logger.info(
+                "Previous run_job_template failed. Retry %d of %d", i, retries
+            )
+        controller_job = await job_template_runner.run_job_template(
+            name, organization, job_args, event_callback
+        )
+        if controller_job["status"] != "failed":
+            break
+
+    await event_log.put(
+        dict(
+            type="Action",
+            action="run_job_template",
+            activation_id=settings.identifier,
+            job_template_name=name,
+            organization=organization,
+            job_id=job_id,
+            ruleset=ruleset,
+            rule=source_rule_name,
+            status=controller_job["status"],
+            run_at=controller_job["created"],
+        )
+    )
+
+    if set_facts or post_events:
+        logger.debug("set_facts")
+        facts = controller_job["artifacts"]
+        if facts:
+            logger.debug("facts %s", facts)
+            if set_facts:
+                lang.assert_fact(ruleset, facts)
+            if post_events:
+                lang.post(ruleset, facts)
+        else:
+            logger.debug("Empty facts are not set")
 
 
 async def shutdown(
@@ -549,6 +772,9 @@ async def shutdown(
     hosts: List,
     variables: Dict,
     facts: Dict,
+    project_data_file: str,
+    source_ruleset_name: str,
+    source_rule_name: str,
     ruleset: str,
 ):
     await event_log.put(
@@ -556,7 +782,10 @@ async def shutdown(
             type="Action",
             action="shutdown",
             activation_id=settings.identifier,
+            ruleset=source_ruleset_name,
+            rule=source_rule_name,
             run_at=str(datetime.utcnow()),
+            matching_events=_get_events(variables),
         )
     )
     raise ShutdownException()
@@ -571,7 +800,9 @@ actions: Dict[str, Callable] = dict(
     post_event=post_event,
     run_playbook=run_playbook,
     run_module=run_module,
+    run_job_template=run_job_template,
     shutdown=shutdown,
+    echo=echo,
 )
 
 
@@ -579,7 +810,7 @@ def update_variables(variables: Dict, var_root: Union[str, Dict]):
     var_roots = {var_root: var_root} if isinstance(var_root, str) else var_root
     if "event" in variables:
         for key, _new_key in var_roots.items():
-            new_value = dpath.util.get(
+            new_value = dpath.get(
                 variables["event"], key, separator=".", default=None
             )
             if new_value:
@@ -588,19 +819,25 @@ def update_variables(variables: Dict, var_root: Union[str, Dict]):
     elif "events" in variables:
         for _k, v in variables["events"].items():
             for old_key, new_key in var_roots.items():
-                new_value = dpath.util.get(
-                    v, old_key, separator=".", default=None
-                )
+                new_value = dpath.get(v, old_key, separator=".", default=None)
                 if new_value:
                     variables["events"][new_key] = new_value
                     break
 
 
-def get_status(private_data_dir: str):
-    status_files = glob.glob(
-        os.path.join(private_data_dir, "artifacts", "*", "status")
-    )
-    status_files.sort(key=os.path.getmtime, reverse=True)
-    with open(status_files[0], "r") as f:
-        status = f.read()
-    return status
+def _get_latest_artifact(data_dir: str, artifact: str, content: bool = True):
+    files = glob.glob(os.path.join(data_dir, "artifacts", "*", artifact))
+    files.sort(key=os.path.getmtime, reverse=True)
+    if content:
+        with open(files[0], "r") as f:
+            content = f.read()
+        return content
+    return files[0]
+
+
+def _get_events(variables: Dict):
+    if "event" in variables:
+        return {"m": variables["event"]}
+    elif "events" in variables:
+        return variables["events"]
+    return {}
