@@ -1,15 +1,26 @@
+#  Copyright 2022 Red Hat, Inc.
+#
+#  Licensed under the Apache License, Version 2.0 (the "License");
+#  you may not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing, software
+#  distributed under the License is distributed on an "AS IS" BASIS,
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#  See the License for the specific language governing permissions and
+#  limitations under the License.
+
 from __future__ import annotations
 
 import asyncio
-import os
+from dataclasses import dataclass
 from typing import Any, Dict, List, NamedTuple, Union
 
-import ansible_rulebook.condition_types as ct
+from drools.ruleset import Ruleset as EngineRuleSet
 
-if os.environ.get("RULES_ENGINE", "drools") == "drools":
-    from drools.ruleset import Ruleset as EngineRuleSet
-else:
-    from durable.lang import ruleset as EngineRuleSet
+import ansible_rulebook.condition_types as ct
 
 
 class EventSourceFilter(NamedTuple):
@@ -52,6 +63,7 @@ class RuleSet(NamedTuple):
 
 class ActionContext(NamedTuple):
     ruleset: str
+    rule: str
     action: str
     actions_args: Dict
     variables: Dict
@@ -63,16 +75,15 @@ class ActionContext(NamedTuple):
 
 class RuleSetQueue(NamedTuple):
     ruleset: RuleSet
-    queue: asyncio.Queue
+    source_queue: asyncio.Queue
 
 
-class RuleSetQueuePlan(NamedTuple):
-    ruleset: RuleSet
+@dataclass
+class Plan:
     queue: asyncio.Queue
-    plan: asyncio.Queue
 
 
 class EngineRuleSetQueuePlan(NamedTuple):
     ruleset: EngineRuleSet
-    queue: asyncio.Queue
-    plan: asyncio.Queue
+    source_queue: asyncio.Queue
+    plan: Plan
