@@ -1,13 +1,27 @@
+#  Copyright 2022 Red Hat, Inc.
+#
+#  Licensed under the Apache License, Version 2.0 (the "License");
+#  you may not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing, software
+#  distributed under the License is distributed on an "AS IS" BASIS,
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#  See the License for the specific language governing permissions and
+#  limitations under the License.
+import asyncio
 import json
-import os
 
 import pytest
 
-from ansible_rulebook.engine import run_rulesets
+from ansible_rulebook.engine import run_rulesets, start_source
+from ansible_rulebook.exception import VarsKeyMissingException
 from ansible_rulebook.messages import Shutdown
 from ansible_rulebook.util import load_inventory
 
-from .test_engine import load_rulebook
+from .test_engine import load_rulebook, validate_events
 
 
 @pytest.mark.asyncio
@@ -28,8 +42,7 @@ async def test_01_noop():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "noop", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "1"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -53,8 +66,7 @@ async def test_02_debug():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "1"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -78,8 +90,7 @@ async def test_03_print_event():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "print_event", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "1"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -103,12 +114,11 @@ async def test_04_set_fact():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "set_fact", "2"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "3"
     assert event["action"] == "print_event", "4"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "5"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "6"
     assert event_log.empty()
 
@@ -131,12 +141,11 @@ async def test_05_post_event():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "post_event", "2"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "3"
     assert event["action"] == "print_event", "4"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "5"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "6"
     assert event_log.empty()
 
@@ -159,6 +168,7 @@ async def test_06_retract_fact():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "set_fact", "2"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "3"
     assert event["action"] == "retract_fact", "4"
@@ -166,8 +176,6 @@ async def test_06_retract_fact():
     assert event["type"] == "Action", "4"
     assert event["action"] == "debug", "5"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
 
@@ -190,8 +198,7 @@ async def test_07_and():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"nested": {"i": 1, "j": 1}}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -215,8 +222,7 @@ async def test_08_or():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"nested": {"i": 1, "j": 1}}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -240,8 +246,7 @@ async def test_09_gt():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"i": 3}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -265,8 +270,7 @@ async def test_10_lt():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"i": 1}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -290,8 +294,7 @@ async def test_11_le():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"i": 2}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -315,8 +318,7 @@ async def test_12_ge():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"i": 2}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -340,8 +342,7 @@ async def test_13_add():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"nested": {"i": 2, "j": 1}}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -365,8 +366,7 @@ async def test_14_sub():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m": {"nested": {"i": 1, "j": 2}}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -390,13 +390,13 @@ async def test_15_multiple_events_all():
         dict(),
     )
 
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "0"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {
+        "m_0": {"nested": {"i": 1, "j": 0}},
+        "m_1": {"nested": {"i": 0, "j": 1}},
+    }, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -422,8 +422,7 @@ async def test_16_multiple_events_any():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {"m_0": {"nested": {"i": 1, "j": 0}}}
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -450,13 +449,11 @@ async def test_17_multiple_sources_any():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "3"
+    assert event["matching_events"] == {"m_0": {"i": 1}}
     event = event_log.get_nowait()
     assert event["type"] == "Action", "4"
     assert event["action"] == "debug", "5"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
+    assert event["matching_events"] == {"m_1": {"range2": {"i": 1}}}, "6"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
@@ -480,15 +477,15 @@ async def test_18_multiple_sources_all():
         dict(),
     )
 
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "0"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
+    assert event["matching_events"] == {
+        "m_0": {"i": 1},
+        "m_1": {"range2": {"i": 1}},
+    }, "3"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "3"
-    event = event_log.get_nowait()
-    assert event["type"] == "Shutdown", "8"
+    assert event["type"] == "Shutdown", "5"
     assert event_log.empty()
 
 
@@ -498,6 +495,7 @@ async def test_19_is_defined():
 
     queue = ruleset_queues[0][1]
     queue.put_nowait(dict(i=1))
+    queue.put_nowait(dict(payload=dict(key1="value1", key2=dict(name="fred"))))
     queue.put_nowait(Shutdown())
 
     await run_rulesets(
@@ -510,11 +508,13 @@ async def test_19_is_defined():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "set_fact", "2"
+    assert event["matching_events"] == {"m": {"i": 1}}
     event = event_log.get_nowait()
     assert event["type"] == "Action", "3"
-    assert event["action"] == "debug", "4"
+    assert event["action"] in ["debug", "print_event"], "4"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "5"
+    assert event["type"] == "Action", "6"
+    assert event["action"] in ["print_event", "debug"], "7"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "6"
     assert event_log.empty()
@@ -538,14 +538,15 @@ async def test_20_is_not_defined():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "set_fact", "2"
+    assert event["matching_events"] == {"m": {"i": 1}}
     event = event_log.get_nowait()
     assert event["type"] == "Action", "3"
     assert event["action"] == "retract_fact", "4"
+    assert event["matching_events"] == {"m": {"msg": "hello"}}
     event = event_log.get_nowait()
     assert event["type"] == "Action", "5"
     assert event["action"] == "debug", "6"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
+    assert event["matching_events"] == {"m": {"msg": "hello"}}
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
@@ -581,8 +582,7 @@ async def test_21_run_playbook(rule):
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "run_playbook", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
+    assert event["matching_events"] == {"m": {"i": 1}}
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
@@ -604,12 +604,9 @@ async def test_23_nested_data():
         dict(),
     )
 
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "0"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "2"
+    assert event["matching_events"] == {"m": {"root": {"nested": {"i": 1}}}}
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "2"
     assert event_log.empty()
@@ -636,8 +633,6 @@ async def test_24_max_attributes():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "2"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "2"
     assert event_log.empty()
 
@@ -665,8 +660,6 @@ async def test_25_max_attributes_nested():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "2"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "2"
     assert event_log.empty()
 
@@ -687,12 +680,10 @@ async def test_26_print_events():
         dict(),
     )
 
-    event_log.get_nowait()
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "print_event", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "1"
+    assert event["matching_events"] == {"m_0": {"i": 1}, "m_1": {"i": 2}}, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -722,23 +713,18 @@ async def test_27_var_root():
         dict(),
     )
 
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "0"
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "print_event", "2"
-
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "6"
+    assert event["matching_events"] == {
+        "webhook": {"url": "http://www.example.com", "action": "merge"},
+        "kafka": {"topic": "testing", "channel": "red"},
+    }
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
 
 
-@pytest.mark.skipif(
-    os.environ.get("RULES_ENGINE", "drools") == "drools",
-    reason="durable rules only test, issues with jinja substitution",
-)
 @pytest.mark.asyncio
 async def test_28_right_side_condition_template():
     ruleset_queues, event_log = load_rulebook(
@@ -757,12 +743,13 @@ async def test_28_right_side_condition_template():
         dict(),
     )
 
-    event_log.get_nowait()
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "1"
+    assert event["matching_events"] == {
+        "first": {"custom": {"expected_index": 2}},
+        "m_1": {"i": 2},
+    }, "3"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "7"
     assert event_log.empty()
@@ -791,12 +778,12 @@ async def test_29_run_module():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "run_module", "2"
-
+    assert event["matching_events"] == {
+        "m": {"i": 1, "meta": {"hosts": "localhost"}}
+    }
     assert event["rc"] == 0, "2.1"
     assert event["status"] == "successful", "2.2"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
 
@@ -830,8 +817,6 @@ async def test_30_run_module_missing():
     assert event["rc"] == 2, "2.1"
     assert event["status"] == "failed", "2.2"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
 
@@ -865,8 +850,6 @@ async def test_31_run_module_missing_args():
     assert event["rc"] == 2, "2.1"
     assert event["status"] == "failed", "2.2"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
 
@@ -900,8 +883,6 @@ async def test_32_run_module_fail():
     assert event["rc"] == 2, "2.1"
     assert event["status"] == "failed", "2.2"
     event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
-    event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
 
@@ -928,10 +909,12 @@ async def test_35_multiple_rulesets_1_fired():
     )
 
     checks = {
-        "max_events": 5,
-        "processed_events": 2,
+        "max_events": 3,
         "shutdown_events": 1,
-        "actions": ["set_fact", "noop"],
+        "actions": [
+            "35 multiple rulesets 1::r1::set_fact",
+            "35 multiple rulesets 1::r2::noop",
+        ],
     }
     validate_events(event_log, **checks)
 
@@ -958,42 +941,16 @@ async def test_36_multiple_rulesets_both_fired():
         dict(),
     )
     checks = {
-        "max_events": 6,
-        "processed_events": 3,
+        "max_events": 3,
         "shutdown_events": 1,
-        "actions": ["set_fact", "debug"],
+        "actions": [
+            "36 multiple rulesets 1::r1::set_fact",
+            "36 multiple rulesets 2::r1::debug",
+        ],
     }
     validate_events(event_log, **checks)
 
 
-def validate_events(event_log, **kwargs):
-    processed_events = 0
-    shutdown_events = 0
-    actions = []
-
-    for _ in range(kwargs["max_events"]):
-        event = event_log.get_nowait()
-        print(event)
-        if event["type"] == "ProcessedEvent":
-            processed_events += 1
-        elif event["type"] == "Action":
-            actions.append(event["action"])
-        elif event["type"] == "Shutdown":
-            shutdown_events += 1
-
-    assert event_log.empty()
-    if "actions" in kwargs:
-        assert kwargs["actions"] == actions
-    if "processed_events" in kwargs:
-        assert kwargs["processed_events"] == processed_events
-    if "shutdown_events" in kwargs:
-        assert kwargs["shutdown_events"] == shutdown_events
-
-
-@pytest.mark.skipif(
-    os.environ.get("RULES_ENGINE", "durable_rules") == "durable_rules",
-    reason="durable rules does not support more than 255 keys in facts",
-)
 @pytest.mark.asyncio
 async def test_37_hosts_facts():
     ruleset_queues, event_log = load_rulebook("examples/37_hosts_facts.yml")
@@ -1012,9 +969,345 @@ async def test_37_hosts_facts():
     event = event_log.get_nowait()
     assert event["type"] == "Action", "1"
     assert event["action"] == "debug", "2"
-
-    event = event_log.get_nowait()
-    assert event["type"] == "ProcessedEvent", "7"
     event = event_log.get_nowait()
     assert event["type"] == "Shutdown", "8"
     assert event_log.empty()
+
+
+@pytest.mark.asyncio
+async def test_38_shutdown_action():
+    ruleset_queues, event_log = load_rulebook("examples/38_shutdown.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=1))
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "shutdown", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "2"
+
+
+@pytest.mark.asyncio
+async def test_40_in():
+    ruleset_queues, event_log = load_rulebook("examples/40_in.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=1))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "3"
+
+
+@pytest.mark.asyncio
+async def test_41_not_in():
+    ruleset_queues, event_log = load_rulebook("examples/41_not_in.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=1))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "3"
+
+
+@pytest.mark.asyncio
+async def test_42_contains():
+    ruleset_queues, event_log = load_rulebook("examples/42_contains.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(id_list=[1, 2, 3]))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "3"
+
+
+@pytest.mark.asyncio
+async def test_43_not_contains():
+    ruleset_queues, event_log = load_rulebook("examples/43_not_contains.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(id_list=[1, 2, 3]))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "3"
+
+
+@pytest.mark.asyncio
+async def test_44_in_and():
+    ruleset_queues, event_log = load_rulebook("examples/44_in_and.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=0))
+    queue.put_nowait(dict(i=1))
+    queue.put_nowait(dict(i=2))
+    queue.put_nowait(dict(i=3))
+    queue.put_nowait(dict(i=4))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "3"
+
+
+@pytest.mark.asyncio
+async def test_45_in_or():
+    ruleset_queues, event_log = load_rulebook("examples/45_in_or.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=4))
+    queue.put_nowait(dict(i=8))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "3"
+    assert event["action"] == "debug", "3"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "5"
+
+
+@pytest.mark.asyncio
+async def test_47_generic_plugin():
+    ruleset_queues, event_log = load_rulebook("examples/47_generic_plugin.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=42))
+    queue.put_nowait(dict(b=True))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "print_event", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "3"
+    assert event["action"] == "print_event", "3"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "5"
+
+
+@pytest.mark.asyncio
+async def test_48_echo():
+    ruleset_queues, event_log = load_rulebook("examples/48_echo.yml")
+
+    queue = ruleset_queues[0][1]
+    queue.put_nowait(dict(i=1))
+    queue.put_nowait(Shutdown())
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "echo", "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "5"
+
+
+@pytest.mark.asyncio
+async def test_49_float():
+    ruleset_queues, event_log = load_rulebook("examples/49_float.yml")
+
+    queue = ruleset_queues[0][1]
+    rs = ruleset_queues[0][0]
+    source_task = asyncio.create_task(
+        start_source(rs.sources[0], ["sources"], {}, queue)
+    )
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "debug", "1"
+    assert event["matching_events"] == {"m": {"pi": 3.14159}}, "3"
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "3"
+    assert event["action"] == "debug", "4"
+    assert event["matching_events"] == {"m": {"mass": 5.97219}}, "5"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "7"
+    source_task.cancel()
+
+
+@pytest.mark.asyncio
+async def test_50_negation():
+    ruleset_queues, event_log = load_rulebook("examples/50_negation.yml")
+
+    queue = ruleset_queues[0][1]
+    rs = ruleset_queues[0][0]
+    source_task = asyncio.create_task(
+        start_source(rs.sources[0], ["sources"], {}, queue)
+    )
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        dict(),
+        load_inventory("playbooks/inventory.yml"),
+    )
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "1"
+    assert event["action"] == "print_event", "1"
+    assert event["matching_events"] == {"m": {"b": False}}, "1"
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "3"
+    assert event["action"] == "print_event", "3"
+    assert event["matching_events"] == {"m": {"bt": True}}, "3"
+    event = event_log.get_nowait()
+    assert event["type"] == "Action", "5"
+    assert event["action"] == "print_event", "5"
+    assert event["matching_events"] == {"m": {"i": 10}}, "5"
+    event = event_log.get_nowait()
+    assert event["type"] == "Shutdown", "7"
+    source_task.cancel()
+
+
+@pytest.mark.asyncio
+async def test_51_vars_namespace():
+    ruleset_queues, event_log = load_rulebook("examples/51_vars_namespace.yml")
+
+    queue = ruleset_queues[0][1]
+    rs = ruleset_queues[0][0]
+    source_task = asyncio.create_task(
+        start_source(rs.sources[0], ["sources"], {}, queue)
+    )
+    years = [2001, 2002, 2003, 2004]
+    address = dict(
+        street="123 Any Street", city="Bedrock", state="NJ", years_active=years
+    )
+    person = dict(
+        person=dict(
+            age=45,
+            name="Fred Flintstone",
+            active=True,
+            reliability=86.9,
+            address=address,
+        )
+    )
+
+    await run_rulesets(
+        event_log,
+        ruleset_queues,
+        person,
+        load_inventory("playbooks/inventory.yml"),
+    )
+    checks = {
+        "max_events": 6,
+        "shutdown_events": 1,
+        "actions": [
+            "51 vars namespace::str_test::echo",
+            "51 vars namespace::list_test::echo",
+            "51 vars namespace::bool_test::echo",
+            "51 vars namespace::int_test::echo",
+            "51 vars namespace::float_test::echo",
+        ],
+    }
+    validate_events(event_log, **checks)
+    source_task.cancel()
+
+
+@pytest.mark.asyncio
+async def test_51_vars_namespace_missing_key():
+    ruleset_queues, event_log = load_rulebook("examples/51_vars_namespace.yml")
+
+    queue = ruleset_queues[0][1]
+    rs = ruleset_queues[0][0]
+    source_task = asyncio.create_task(
+        start_source(rs.sources[0], ["sources"], {}, queue)
+    )
+    years = [2001, 2002, 2003, 2004]
+    address = dict(
+        street="123 Any Street", city="Bedrock", state="NJ", years_active=years
+    )
+    person = dict(
+        person=dict(
+            name="Fred Flintstone",
+            active=True,
+            reliability=86.9,
+            address=address,
+        )
+    )
+
+    with pytest.raises(VarsKeyMissingException) as exc_info:
+        await run_rulesets(
+            event_log,
+            ruleset_queues,
+            person,
+            load_inventory("playbooks/inventory.yml"),
+        )
+    assert str(exc_info.value) == "vars does not contain key: person.age"
+    source_task.cancel()
